/*
 * File:   graph.h
 * Author: Filip Niksic, fniksic@gmail.com
 *
 * Created on January 23, 2008, 10:50 PM
 */

#ifndef GRAPH_H
#define	GRAPH_H

#include <cstddef>
#include <vector>
#include <set>
#include <stdexcept>
#include <iostream>
#include <utility>

#include "globals.h"

class Graph; // forward deklaracija

/* Klasa Graph_data mi predstavlja enkapsulaciju grafa. Ona se nece direktno instancirati,
 * nego ce se njome baratati preko handle klase Graph.
 */
class Graph_data {
    friend class Graph;
    friend std::ostream &operator<<(std::ostream &, const Graph &);
    
    typedef std::vector<Real_T> Row;
    typedef std::vector<Row> Matrix;
    typedef std::size_t Vertex_ID;
    typedef std::pair<Vertex_ID, Vertex_ID> Edge_ID;
    
    Vertex_ID size; // broj vrhova
    Matrix adjMatrix; // matrica susjedstva
    
    static const Vertex_ID nullVertex;
    static const Edge_ID nullEdge;
    
    /* Konstruktor koji prima broj vrhova grafa i inicijalizira matricu susjedstva.
     * Matrica susjedstva se inicijalno postavlja tako da svuda bude Inf, tj.
     * ne postoji nijedan brid u grafu.
     *
     * Konstruktor je eksplicitan da izbjegnemo nezeljene implicitne konverzije
     * size_t -> Graph_data
     */
    explicit Graph_data(Vertex_ID n = 0) : size(n), adjMatrix(n, Row(n, Inf)) {}
    
    /* Set funkcija za baratanje bridovima */
    
    void setEdge(Vertex_ID u, Vertex_ID v, Real_T value) {
        validateVertex(u);
        validateVertex(v);
        adjMatrix[u][v] = adjMatrix[v][u] = value;
    }
    void setEdge(Edge_ID e, Real_T value) {
        setEdge(e.first, e.second, value);
    }
    void resetEdge(Vertex_ID u, Vertex_ID v) {
        validateVertex(u);
        validateVertex(v);
        setEdge(u, v, Inf);
    }
    void resetEdge(Edge_ID e) {
        resetEdge(e.first, e.second);
    }
    bool existsEdge(Vertex_ID u, Vertex_ID v) const {
        validateVertex(u);
        validateVertex(v);
        return adjMatrix[u][v] != Inf;
    }
    bool existsEdge(Edge_ID e) const {
        return existsEdge(e.first, e.second);
    }
    Real_T getEdge(Vertex_ID u, Vertex_ID v) const {
        validateVertex(u);
        validateVertex(v);
        return adjMatrix[u][v];
    }
    Real_T getEdge(Edge_ID e) const {
        return getEdge(e.first, e.second);
    }
    /* clearEdges() postavlja sve vrijednosti u adjMatrix na Inf */
    void clearEdges();
    
    /* Funkcije za baratanje vrhovima */
    
    Vertex_ID addVertex();
    void delVertex(Vertex_ID);
    
    Vertex_ID numVertices() const {
        return size;
    }
    
    void validateVertex(Vertex_ID v) const {
        if (v < 0 || v >= size)
            throw std::out_of_range("Nepostojeci vrh.");
    }
    
    Graph_data *subGraph(const std::set<Vertex_ID> &) const;
    
    void print(std::ostream &) const;
};

class Graph {
    friend std::ostream &operator<<(std::ostream &, const Graph &);
    Graph_data *ptr;
    size_t *count;
public:
    typedef Graph_data::Vertex_ID Vertex_ID;
    typedef Graph_data::Edge_ID Edge_ID;
    
    static const Vertex_ID nullVertex;
    static const Edge_ID nullEdge;
    
    explicit Graph(Vertex_ID n = 0) : ptr(new Graph_data(n)), count(new size_t(1)) {}
    
    Graph(Graph_data *_ptr) : ptr(_ptr), count(new size_t(1)) {}
    
    // Copy control
    
    Graph(const Graph &g) : ptr(g.ptr), count(g.count) {
        ++*count;
    }
    ~Graph() {
        decr_count();
    }
    Graph &operator=(const Graph &g) {
        ++*g.count;
        decr_count();
        ptr = g.ptr;
        count = g.count;
        return *this;
    }
    
    void setEdge(Vertex_ID u, Vertex_ID v, Real_T value) {
        ptr->setEdge(u, v, value);
    }
    void setEdge(Edge_ID e, Real_T value) {
        setEdge(e.first, e.second, value);
    }
    void resetEdge(Vertex_ID u, Vertex_ID v) {
        ptr->resetEdge(u, v);
    }
    void resetEdge(Edge_ID e) {
        resetEdge(e.first, e.second);
    }
    bool existsEdge(Vertex_ID u, Vertex_ID v) const {
        return ptr->existsEdge(u, v);
    }
    bool existsEdge(Edge_ID e) const {
        return existsEdge(e.first, e.second);
    }
    Real_T getEdge(Vertex_ID u, Vertex_ID v) const {
        return ptr->getEdge(u, v);
    }
    Real_T getEdge(Edge_ID e) const {
        return getEdge(e.first, e.second);
    }
    void clearEdges() {
        ptr->clearEdges();
    }
    Vertex_ID addVertex() {
        return ptr->addVertex();
    }
    void delVertex(Vertex_ID v) {
        ptr->delVertex(v);
    }
    Vertex_ID numVertices() const {
        return ptr->numVertices();
    }
    Graph_data *subGraph(const std::set<Vertex_ID> &subset) const {
        return ptr->subGraph(subset);
    }
private:
    void decr_count() {
        if (--*count == 0) {
            delete ptr;
            delete count;
        }
    }
};

#endif	/* GRAPH_H */

